import logging
import unittest
from unittest import mock

from pika.exceptions import AMQPConnectionError, NackError, UnroutableError

from sdc.rabbit import QueuePublisher
from sdc.rabbit.exceptions import PublishMessageError
from sdc.rabbit.test.test_data import test_data


class TestPublisher(unittest.TestCase):
    logger = logging.getLogger(__name__)

    publisher = QueuePublisher(['amqp://guest:guest@0.0.0.0:5672',
                                'amqp://guest:guest@0.0.0.0:5672'],
                               'test')

    bad_publisher = QueuePublisher(['amqp://guest:guest@0.0.0.0:672',
                                    'amqp://guest:guest@0.0.0.0:672'],
                                   'test')

    confirm_delivery_publisher = QueuePublisher(['amqp://guest:guest@0.0.0.0:5672',
                                                 'amqp://guest:guest@0.0.0.0:5672'],
                                                'test',
                                                confirm_delivery=True)

    def test_init(self):
        this_publisher = QueuePublisher(['amqp://guest:guest@0.0.0.0:5672',
                                         'amqp://guest:guest@0.0.0.0:5672'],
                                        'test')
        self.assertEqual(this_publisher._urls, ['amqp://guest:guest@0.0.0.0:5672',
                                                'amqp://guest:guest@0.0.0.0:5672'])
        self.assertEqual(this_publisher._queue, 'test')
        self.assertEqual(this_publisher._arguments, {})
        self.assertEqual(this_publisher._connection, None)
        self.assertEqual(this_publisher._channel, None)

    def test_connect_loops_correctly(self):
        this_publisher = QueuePublisher(['amqp://guest:guest@0.0.0.0:672',
                                         'amqp://guest:guest@0.0.0.0:5672'],
                                        'test')

        self.assertEqual(this_publisher._urls, ['amqp://guest:guest@0.0.0.0:672',
                                                'amqp://guest:guest@0.0.0.0:5672'])
        self.assertEqual(this_publisher._queue, 'test')
        self.assertEqual(this_publisher._arguments, {})
        self.assertEqual(this_publisher._connection, None)
        self.assertEqual(this_publisher._channel, None)

        this_publisher._connect()

    def test_connect_amqp_connection_error(self):
        with self.assertRaises(AMQPConnectionError):
            self.bad_publisher._connect()

    def test_connect_confirm_delivery_true(self):
        with self.assertLogs(level='INFO') as cm:
            self.confirm_delivery_publisher._connect()

        msg = 'Enabled delivery confirmation'
        self.assertIn(msg, cm[0][3].msg)

    def test_connect_amqpok(self):

        result = self.publisher._connect()
        self.assertEqual(result, True)

    def test_disconnect_ok(self):
        self.publisher._connect()

        with self.assertLogs(level='DEBUG') as cm:
            self.publisher._disconnect()

        msg = 'Disconnected from queue'
        self.assertIn(msg, cm[1][-1])

    def test_disconnect_error(self):

        self.publisher._disconnect()
        with self.assertLogs(level='ERROR') as cm:
            self.publisher._disconnect()

        msg = 'Unable to close connection'
        self.assertIn(msg, cm.output[0])

    def test_publish_message_no_connection(self):
        with self.assertRaises(PublishMessageError):
            self.bad_publisher.publish_message(test_data['valid'])

    def test_publish(self):
        self.publisher._connect()
        result = self.publisher.publish_message(test_data['valid'])
        self.assertEqual(True, result)

    def test_publish_nack_error(self):
        mock_method = 'pika.adapters.blocking_connection.BlockingChannel.basic_publish'
        with mock.patch(mock_method) as barMock:
            barMock.side_effect = NackError('a')
            self.publisher._connect()
            with self.assertRaises(PublishMessageError):
                self.publisher.publish_message(test_data['valid'])

    def test_publish_unroutable_error(self):
        mock_method = 'pika.adapters.blocking_connection.BlockingChannel.basic_publish'
        with mock.patch(mock_method) as barMock:
            barMock.side_effect = UnroutableError('a')
            self.publisher._connect()
            with self.assertRaises(PublishMessageError):
                self.publisher.publish_message(test_data['valid'])

    def test_publish_generic_error(self):
        mock_method = 'pika.adapters.blocking_connection.BlockingChannel.basic_publish'
        with mock.patch(mock_method) as barMock:
            barMock.side_effect = Exception()
            self.publisher._connect()
            with self.assertRaises(Exception):
                self.publisher.publish_message(test_data['valid'])



import logging

import pika
from pika.exceptions import NackError, UnroutableError

from sdc.rabbit.exceptions import PublishMessageError

logger = logging.getLogger(__name__)


class QueuePublisher(object):
    """This is a queue publisher that publishes response messages to a
    RabbitMQ queue.

    """
    _durable_queue = True

    def __init__(self, urls, queue, **kwargs):
        """Create a new instance of the QueuePublisher class

        :param logger: A reference to a logging.Logger instance
        :param urls: List of RabbitMQ cluster URLs.
        :param queue: Queue name
        :param confirm_delivery: Delivery confirmations toggle
        :param **kwargs: Custom key/value pairs passed to the arguments
            parameter of pika's channel.queue_declare method

        :returns: Object of type QueuePublisher
        :rtype: QueuePublisher

        """
        self._urls = urls
        self._queue = queue
        self._arguments = kwargs
        self._connection = None
        self._channel = None
        self._confirm_delivery = False
        if 'confirm_delivery' in kwargs:
            self._confirm_delivery = True
            self._arguments.pop('confirm_delivery', None)

    def _connect(self):
        """
        Connect to a RabbitMQ queue

        :returns: Boolean corresponding to success of connection
        :rtype: bool

        """
        logger.info("Connecting to queue")
        for url in self._urls:
            try:
                self._connection = pika.BlockingConnection(pika.URLParameters(url))
                self._channel = self._connection.channel()
                self._channel.queue_declare(queue=self._queue,
                                            durable=self._durable_queue,
                                            arguments=self._arguments)
                if self._confirm_delivery:
                    self._channel.confirm_delivery()
                    logger.info("Enabled delivery confirmation")
                logger.debug("Connected to queue")
                return True

            except pika.exceptions.AMQPConnectionError:
                logger.exception("Unable to connect to queue")
                continue

        raise pika.exceptions.AMQPConnectionError

    def _disconnect(self):
        """
        Cleanly close a RabbitMQ queue connection.

        :returns: None

        """
        try:
            self._connection.close()
            logger.debug("Disconnected from queue")

        except Exception:
            logger.exception("Unable to close connection")

    def publish_message(self, message, content_type=None, headers=None, mandatory=False, immediate=False):
        """
        Publish a response message to a RabbitMQ queue.

        :param message: Response message
        :param content_type: Pika BasicProperties content_type value
        :param headers: Message header properties
        :param mandatory: The mandatory flag
        :param immediate: The immediate flag

        :returns: Boolean corresponding to the success of publishing
        :rtype: bool

        """
        logger.debug("Publishing message")
        try:
            self._connect()
        except pika.exceptions.AMQPConnectionError:
            logger.error("Message not published. RetryableError raised")
            raise PublishMessageError

        try:
            result = self._channel.basic_publish(exchange='',
                                                 routing_key=self._queue,
                                                 mandatory=mandatory,
                                                 immediate=immediate,
                                                 properties=pika.BasicProperties(
                                                     content_type=content_type,
                                                     headers=headers,
                                                     delivery_mode=2
                                                 ),
                                                 body=message)

            logger.info('Published message to queue', queue=self._queue)
            return result
        except NackError:
            # raised when a message published in publisher-acknowledgments mode
            # is returned via `Basic.Return` followed by `Basic.Ack`.
            logger.error("NackError occured. Message not published.")
            raise PublishMessageError
        except UnroutableError:
            # raised when a message published in publisher-acknowledgments
            # mode is returned via `Basic.Return` followed by `Basic.Ack`.
            logger.error("UnroutableError occured. Message not published.")
            raise PublishMessageError
        except Exception:
            logger.exception("Unknown exception occured. Message not published.")
            raise PublishMessageError
